module HAL where

{-# LANGUAGE BangPatterns #-}

import Conduit
import Control.Monad
import Control.Monad.Reader
import Data.Aeson
import Data.Default (def)
import Data.Either (fromRight)
import Data.Maybe (fromMaybe)
import Data.Text
import HAL.Client
import HAL.Doc.Corpus
import HAL.Doc.Struct
import Network.HTTP.Client (newManager, Request)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import Prelude
import Servant.API
import Servant.Client (BaseUrl(..), Scheme(..), ClientM, ClientError, runClientM, mkClientEnv, ClientEnv (makeClientRequest))
import System.IO.Unsafe (unsafePerformIO)

data HalCrawlerOptions
  = HalCrawlerOptions
  { -- | If 'True', enable the debug logs to stdout.
    _hco_debugLogs :: !Bool
  , _hco_batchSize :: !Int
  }

defaultHalOptions :: HalCrawlerOptions
defaultHalOptions = HalCrawlerOptions
  { _hco_debugLogs = False
  , _hco_batchSize = 1000
  }

getMetadataWithOptionsC :: HalCrawlerOptions
                        -- ^ The options for the crawler
                        -> Text
                        -- ^ The textual query
                        -> Maybe Int
                        -- ^ An optional offset
                        -> Maybe Int
                        -- ^ An optional limit for the search, it influences the number of
                        -- rows returned.
                        -> IO (Either ClientError (Maybe Int, ConduitT () Corpus IO ()))
getMetadataWithOptionsC opts@HalCrawlerOptions{..} q mb_offset mb_limit = do
  manager' <- newManager tlsManagerSettings
  -- First, estimate the total number of documents
  eCount <- countResults q
  pure $ get' <$> eCount
  where
    get' :: Int -> (Maybe Int, ConduitT () Corpus IO ())
    get' numFound =
      ( Just numResults
      , yieldMany [0..]
        .| takeC numPages
        .| concatMapMC (getPage q offset))
      where
        offset = fromMaybe 0 mb_offset
        limit  = min numFound $ fromMaybe numFound mb_limit
        numResults = limit - offset
        numPages = numResults `div` _hco_batchSize + 1

    getPage :: Text -> Int -> Int -> IO [Corpus]
    getPage q start pageNum = do
      let offset = start + pageNum * _hco_batchSize
      debugLog opts $ "[getMetadataWithC] getPage: " <> show offset
      eRes <- runHalAPIClient opts $ search (Just requestedFields) [q] Nothing (Just offset) (Just _hco_batchSize)
      pure $ case eRes of
        Left _ -> []
        Right (Response { _docs }) -> _docs

    printDoc :: Corpus -> IO Corpus
    printDoc c@(Corpus { _corpus_docid, _corpus_title }) = do
      debugLog opts $ show _corpus_title
      pure c

debugLog :: HalCrawlerOptions -> String -> IO ()
debugLog HalCrawlerOptions{..} msg = when _hco_debugLogs $ putStrLn msg

getMetadataWith :: Text
                -- ^ The query, as a text.
                -> Maybe Int
                -- ^ The offset, it influences where the search starts.
                -> Maybe Int
                -- ^ The offset, it influences how many rows are returned.
                -> IO (Either ClientError (Response Corpus))
getMetadataWith q start rows = do
  manager' <- newManager tlsManagerSettings
  runHalAPIClient defaultHalOptions $ search (Just requestedFields) [q] Nothing start rows

getMetadataWithC :: Text
                 -- ^ The textual query
                 -> Maybe Int
                 -- ^ An optional offset for the search, it influences where to start.
                 -> Maybe Int
                 -- ^ An optional limit for the search, it influences how many rows are
                 -- returned.
                 -> IO (Either ClientError (Maybe Int, ConduitT () Corpus IO ()))
getMetadataWithC q start = getMetadataWithOptionsC defaultHalOptions q start

countResults :: Text -> IO (Either ClientError Int)
countResults q = do
  manager' <- newManager tlsManagerSettings
  -- First, estimate the total number of documents
  eRes <- runHalAPIClient defaultHalOptions $ search (Just requestedFields) [q] Nothing (Just 0) (Just 1) :: IO (Either ClientError (Response Corpus))
  pure $ _numFound <$> eRes

requestedFields :: Text
requestedFields = "docid,title_s,abstract_s,submittedDate_s,source_s,authFullName_s,authOrganism_s"

structFields :: Text
structFields = "docid,label_s,parentDocid_i"

requestLog :: HalCrawlerOptions -> Request -> Request
requestLog opts rq = unsafePerformIO $ do
  debugLog opts $ "[HAL.makeClientRequestLog] " <> show rq
  pure rq
{-# NOINLINE requestLog #-}

runHalAPIClient :: (FromJSON doc, ToHttpApiData doc)
                => HalCrawlerOptions
                -> ClientM (Response doc)
                -> IO (Either ClientError (Response doc))
runHalAPIClient opts cmd = do
  manager' <- newManager tlsManagerSettings
  runClientM cmd' (mkClientEnv manager' $ BaseUrl Https "api.archives-ouvertes.fr" 443 "")
  where
    cmd' = local (\r -> r {
      makeClientRequest = \bUrl servantRq -> requestLog opts ((makeClientRequest r) bUrl servantRq)
      }) cmd

runStructureRequest :: Maybe Text -> IO (Either ClientError (Response Struct))
runStructureRequest rq =
  runHalAPIClient defaultHalOptions $ structure (Just structFields) rq (Just 10000)

runSearchRequest :: [Text] -> IO (Either ClientError (Response Corpus))
runSearchRequest rq =
  runHalAPIClient defaultHalOptions $ search (Just requestedFields) rq Nothing Nothing Nothing

generateRequestByStructID :: Text -> [Text] -> Text
generateRequestByStructID rq struct_ids =
  "text:" <>
  rq
  <> " AND "
  <> "structId_i:("
  <> (flattenPipe struct_ids)
  <> ")"

flattenPipe :: [Text] -> Text
flattenPipe (x:[]) = x
flattenPipe (x:xs) = x <> " || " <> flattenPipe xs
